/*
 * Copyright (c) 1992-1993 Michael A. Cooper.
 * This software may be freely distributed provided it is not sold for 
 * profit and the author is credited appropriately.
 */

/*
 * $Header: /src/common/usc/bin/sysinfo/RCS/defs.h,v 1.56 1993/06/06 22:46:19 mcooper Exp mcooper $
 */

#ifndef __sysinfo_defs__
#define __sysinfo_defs__ 

#include <stdio.h>
#include "os.h"
#include "options.h"		/* Need to include options.h before errno.h */
#include <errno.h>
#include "version.h"
#include "config.h"

#if defined(HAVE_UNAME)
#	include <sys/utsname.h>
#endif	/* HAVE_UNAME */
#if defined(HAVE_SYSINFO)
#	include <sys/systeminfo.h>
#endif	/* HAVE_SYSINFO */

#if defined(NEED_KVM)
#	include "kvm.h"
#else
#if defined(HAVE_KVM)
#	include <kvm.h>
#endif	/* HAVE_KVM */
#endif	/* NEED_KVM */

#if	defined(HAVE_MEMSET)
#	define BCOPY(a,b,s)	memcpy(b,a,s)
#	define BZERO(a,s)	memset(a,0,s)
#else
#	define BCOPY(a,b,s)	bcopy(a,b,s)
#	define BZERO(a,b)	bzero(a,b)
#endif	/* HAVE_MEMSET */
#if	defined(HAVE_STRCHR)
#	define index		strchr
#	define rindex		strrchr
#endif	/* HAVE_STRCHR */

/*
 * File containing our CPU model name.  Overrides all other methods
 * for determing model name.
 */
#if !defined(MODELFILE)
#	define MODELFILE	"/etc/sysmodel"
#endif

/*
 * System call failure value
 */
#if !defined(SYSFAIL)
#	define SYSFAIL		-1
#endif

#define C_NULL			'\0'

/*
 * System error numbers and message strings
 */
extern int errno;
#if defined(SYSERR)
#	undef SYSERR
#endif
#define SYSERR			strerror(errno)

#if !defined(MAXHOSTNAMLEN)
#	define MAXHOSTNAMLEN 	256
#endif

#define MHERTZ			1000000			/* MegaHertz */
#define MBYTES			1048576			/* Megabytes */
#define KBYTES			1024			/* Kilobytes */

/*
 * Misc macros
 */
#define EQ(a,b)			(strcasecmp(a,b)==0)
#define EQN(a,b,n)		(strncasecmp(a,b,n)==0)
#define ARG(s)			((s) ? s : "<null>")

/*
 * CheckNlist() breaks in a variety of ways on
 * various OS's.
 */
#if !defined(BROKEN_NLIST_CHECK)
#	define CheckNlist(p) 	_CheckNlist(p)
#else
#	define CheckNlist(p) 	0
#endif

/*
 * Get nlist n_name
 */
#if defined(__MACH__)
#	define GetNlName(n)	(n.n_un.n_name)
#	define GetNlNamePtr(n)	(n->n_un.n_name)
#else
#	define GetNlName(n)	(n.n_name)
#	define GetNlNamePtr(n)	(n->n_name)
#endif	/* __MACH__ */

/*
 * Conversion macros
 */
#define bytes_to_mbytes(n)	((float) n / (float) 1048576)
#define bytes_to_kbytes(n)	((u_long)(n / 1024))
#define mbytes_to_bytes(n)	((u_long)((u_long) n * 1048576))
#define kbytes_to_mbytes(n)	((u_long)((u_long) n / 1024))
#define nsect_to_bytes(n,s)  	((((float) n) / (float) (1024 / s)) * 1024)

/*
 * Are flags f set in b?
 */
#define FLAGS_ON(b,f)		((b != 0) && (b & f))

/*
 * Verbosity levels
 */
#define L_BRIEF			0x01		/* Briefest */
#define L_GENERAL		0x02		/* General verbosity */
#define L_DESC			0x04		/* Description info */
#define L_CONFIG		0x10		/* Configuratin info */
#define L_DEBUG			0x20		/* Debug info */
#define L_ALL (L_BRIEF|L_GENERAL|L_DESC|L_CONFIG)
/*
 * Verbosity macros
 */
#define VL_BRIEF		(Level & L_BRIEF)
#define VL_GENERAL		(Level & L_GENERAL)
#define VL_DESC			(Level & L_DESC)
#define VL_CONFIG		(Level & L_CONFIG)
#define VL_ALL			(Level == L_ALL)
#define VL_DEBUG		(Level & L_DEBUG)

/*
 * OBP node ID type
 */
typedef int			OBPnodeid_t;

/*
 * Show info
 */
struct ShowInfo {
    int				Type;
    char		       *Name;
    char		       *Label;
    void		      (*ShowFunc)();
    char		     *(*GetFunc)();
    int			       *OptCompat;
    int				Enable;
};
typedef struct ShowInfo SHOWINFO;

/*
 * ShowInfo types (ShowInfo.Type)
 */
#define S_GEN			1		/* General info */
#define S_DEV			2		/* Device info */

/*
 * Main Device information
 *
 * Used after device info has been obtained 
 */
struct _device {
    char		*dv_name;		/* Name (e.g. cgtwo0) */
    char		*dv_aname;		/* Alt name */
    char	       **dv_files;		/* Device files */
    int			 dv_type;		/* Device type (eg DT_TAPE) */
    char		*dv_model;		/* Model */
    char		*dv_desc;		/* Description */
    char	       **dv_desclist;		/* Description List */
    int			 dv_unit;		/* Unit number */
    int			 dv_addr;		/* Address */
    int			 dv_prio;		/* Priority */
    int			 dv_vec;		/* Vector */
    int			 dv_nodeid;		/* ID of this node */
    char		*dv_mastername;		/* Name of master */
    caddr_t		*dv_devspec;		/* Device specific info */
    struct _device	*dv_master;		/* Device controller */
    struct _device	*dv_slaves;		/* Devices on this device */
    struct _device	*dv_nxt;		/* Pointer to next device */
};
typedef struct _device DEVICE;
#define MAXDESCLIST	 3			/* Dflt max no of desclist */

/*
 * Device types (DEVICE.dv_type)
 */
#define DT_NONE			 1		/* Sort of ignore */
#define DT_GENERIC		 2		/* Generic Device */
#define DT_DISKDRIVE		 3		/* Disk Drive */
#define DT_DISKCTLR		 4		/* Disk Controller */
#define DT_TAPEDRIVE		 5		/* Tape Drive */
#define DT_TAPECTLR		 6		/* Tape Controller */
#define DT_FRAMEBUFFER		 7		/* Frame Buffer */
#define DT_NETIF		 8		/* Network Interface */
#define DT_BUS			 9		/* System Bus */
#define DT_PSEUDO		10		/* Pseudo Device */
#define DT_CPU			11		/* CPU */
#define DT_MEMORY		12		/* Memory */

/*
 * Disk type
 */
#define DKT_GENERIC		1		/* Generic disk */
#define DKT_CDROM		2		/* CD-ROM */

/*
 * Disk Partition information.
 */
struct _dkpart {
    char		*dp_name;		/* Partition name */
    char		*dp_mnt;		/* Mount information */
    int			 dp_stsect;		/* Starting sector */
    int			 dp_nsect;		/* Number of sectors */
    struct _dkpart	*dp_nxt;		/* Pointer to next DISKPART */
};
typedef struct _dkpart DISKPART;

/*
 * Disk Drive specific data
 */
struct _dkdrive {
    char		*dd_label;		/* Disk label */
    int			 dd_unit;		/* Unit number */
    int			 dd_slave;		/* Slave number */
    int			 dd_dcyl;		/* # data cylinders */
    int			 dd_pcyl;		/* # physical cylinders */
    int			 dd_acyl;		/* # alternate cylinders */
    int			 dd_heads;		/* Number of heads */
    int			 dd_sect;		/* Number of sectors */
    int			 dd_psect;		/* Number of physical sector */
    int			 dd_promrev;		/* PROM Revision */
    int			 dd_apc;		/* Alternates / Cyl (SCSI) */
    int			 dd_rpm;		/* Revolutions Per Minute */
    int			 dd_intrlv;		/* Interleave factor */
    int			 dd_secsize;		/* Size of Sector (bytes) */
    u_long		 dd_size;		/* Size of disk in bytes */
    int			 dd_flags;		/* Info flags */
    DEVICE		*dd_ctlr;		/* Controller disk is on */
    struct _dkpart	*dd_part;		/* Partition information */
    struct _dkdrive	*dd_nxt;		/* Pointer to next disk */
};
typedef struct _dkdrive DISKDRIVE;

/*
 * Disk Flags (DISKDRIVE.dd_flags)
 */
#define DF_HEXUNIT	0x01			/* Unit is prt 3 hex digits */

/*
 * Generic Name/Value table
 */
struct _nametab {
    int			value;			/* Value field */
    char       	       *name;			/* Corresponding name */
    char       	       *valuestr;		/* Value string field */
};
typedef struct _nametab NAMETAB;

/*
 * Generic Key/Value table
 */
struct _keytab {
    long		kt_lvalue;		/* Long value field */
    char       	       *kt_key;			/* Corresponding name */
    char       	       *kt_svalue;		/* Value string field */
};
typedef struct _keytab KEYTAB;

/*
 * Part information
 */
struct _partinfo {
    char       	       *pt_part;		/* Part number/string */
    char       	       *pt_name;		/* Part name */
    char       	       *pt_desc;		/* Part description */
};
typedef struct _partinfo PARTINFO;

/*
 * Disk Controller Table
 */
struct dkctlrtab {
    u_short		 ct_ctype;		/* Ctlr type index */
    char		*ct_model;		/* Ctlr model name */
    int			 ct_rdcmd;		/* Read command flag */
};
typedef struct dkctlrtab DKCTLRTAB;

/*
 * FrameBuffer specific data
 */
struct _framebuffer {
    int			 fb_height;		/* Height (in pixels) */
    int			 fb_width;		/* Width (in pixels) */
    int			 fb_depth;		/* Depth (bits/pixel) */
    u_long		 fb_size;		/* Total size (in bytes) */
    u_long		 fb_vmsize;		/* Video memory (in bytes) */
    int			 fb_cmsize;		/* Color Map Size (#entries) */
};
typedef struct _framebuffer FRAMEBUFFER;

/*
 * Network Interface specific data
 */
struct _netif {
    char		*ni_typename;		/* Name of address type */
    char		*ni_hostaddr;		/* Host address */
    char		*ni_hostname;		/* Host name */
    char		*ni_macaddr;		/* Current MAC address */
    char		*ni_macname;		/* Current MAC name */
    char		*ni_fmacaddr;		/* Factory MAC address */
    char		*ni_fmacname;		/* Factory MAC name */
    char		*ni_netaddr;		/* Network address */
    char		*ni_netname;		/* Network name */
    struct _netif	*ni_nxt;		/* Pointer to next element */
};
typedef struct _netif NETIF;

/*
 * Address family table
 */
struct _aftab {
    int			af_type;		/* Type value */
    char	       *af_name;		/* Name value */
    NETIF	     *(*af_getnetif)();		/* Function to get netif */
};
typedef struct _aftab AFTAB;

/*
 * Table of devices
 */
struct devicetab {
    int			 d_type;		/* Device type */
    char	       **d_list;		/* List of devices */
    DEVICE	      *(*d_query)();		/* Query function */
};

/*
 * Basic device data
 *
 * Used when searching for initial list of devices
 */
struct _devdata {
    char		*dd_devname;		/* Name of device */
    int			 dd_devunit;		/* Device specific unit # */
    int			 dd_slave;		/* Slave number */
    dev_t		 dd_devnum;		/* Device number */
    char		*dd_ctlrname;		/* Name of Controller */
    int			 dd_ctlrunit;		/* Controller # */
    int			 dd_flags;		/* Device flags */
    int			 dd_nodeid;		/* ID of this node */
};
typedef struct _devdata DEVDATA;

/*
 * Flags for DEVDATA.dd_flags
 */
#define DD_MAYBE_ALIVE	0x1			/* Device may be alive */
#define DD_IS_ALIVE	0x2			/* Device is alive */

/*
 * Device Data Table structure
 */
struct _devdatatab {
    char		*ddt_name;		/* Name of device */
    char		*ddt_model;		/* Model */
    char		*ddt_desc;		/* Description */
    DEVICE	      *(*ddt_probe)();		/* Probe device */
    int			 ddt_type;		/* Type of device */
    int			 ddt_flags;		/* Flags */
};
typedef struct _devdatatab DEVDATATAB;

/*
 * Flags for DEVDATATAB.ddt_flags
 */
#define DDT_LENCMP	0x1			/* Compare by length */
#define DDT_NOUNIT	0x2			/* No unit number */

/*
 * Declarations
 */
extern int 			DoPrintUnknown;
extern int 			Debug;
extern int 			Level;
extern int 			Terse;
extern int 			UseProm;

char 			       *index();
char		 	       *rindex();
char 			       *strdup();
char 			       *strcat();
char 			       *strtok();

char			       *itoa();
char	 		       *xmalloc();
char	 		       *xrealloc();
char	 		       *xcalloc();

extern DEVDATATAB	       *GetDevDataTab();
extern DEVICE	 	       *NewDevice();
extern DEVICE	 	       *OBPprobeCPU();
extern DEVICE	 	       *ProbeDiskDrive();
extern DEVICE	 	       *ProbeFrameBuffer();
extern DEVICE	 	       *ProbeGeneric();
extern DEVICE	 	       *ProbeNetif();
extern DEVICE	 	       *ProbeTapeDrive();
extern DEVICE 		       *FindDeviceByName();
extern DEVICE 		       *FindDeviceByNodeID();
extern DEVICE 		       *MkMasterFromDevData();
extern DEVICE 		       *ProbeDevice();
extern DEVICE 		       *ProbeUnknown();
extern DISKDRIVE               *NewDiskDrive();
extern DISKPART  	       *NewDiskPart();
extern DKCTLRTAB 	       *GetDkCtlrTab();
extern FRAMEBUFFER 	       *NewFrameBuffer();
extern NAMETAB	 	       *GetFBTab();
extern NETIF	 	       *NewNetif();
extern PARTINFO	 	       *GetPartInfo();
extern char		       *FreqStr();
extern char		       *GetCpuName();
extern char		       *OBPGetCpuType();
extern char		       *GetHostName();
extern char		       *GetKernArch();
extern char		       *GetKernArchName();
extern char		       *GetKernelVersionStr();
extern char		       *GetModelName();
extern char		       *GetNameTabName();
extern char		       *GetOSNameStr();
extern char		       *GetOSVersionStr();
extern char		       *GetSerialNoStr();
extern char		       *MkDevName();
extern char		       *RunCmds();
extern char		       *RunTestFiles();
extern char	 	       *GetMemoryFromPhysmem();
extern char	 	       *MkDevName();
extern char 		       *GetAppArch();	
extern char 		       *GetCPU();
extern char 		       *GetCharFile();
extern char 		       *GetDiskCapacity();
extern char 		       *GetHostAddrs();
extern char 		       *GetHostAliases();
extern char 		       *GetHostID();
extern char 		       *GetKernelVersion();
extern char 		       *GetMan();
extern char 		       *GetManLong();
extern char 		       *GetManShort();
extern char 		       *GetMemory();
extern char 		       *GetModel();
extern char 		       *GetOSName();
extern char 		       *GetOSVersion();
extern char 		       *GetRawFile();
extern char 		       *GetRomVer();
extern char 		       *GetSerialNo();
extern char 		       *GetVirtMem();
extern int		        AddDevice();
extern void			Error();
extern void		        SetMacInfo();
extern void		        ShowDevices();
extern void 		        ShowGeneral();

#if	defined(HAVE_KVM)
extern kvm_t 		       *KVMopen();
extern void			KVMclose();
extern int			KVMread();
extern struct nlist	       *KVMnlist();
#endif	/* HAVE_KVM */

#endif /* __sysinfo_defs__ */
