/*
 * Copyright (c) 1992-1993 Michael A. Cooper.
 * This software may be freely distributed provided it is not sold for 
 * profit and the author is credited appropriately.
 */

#ifndef lint
static char *RCSid = "$Header: /src/common/usc/bin/sysinfo/RCS/macinfo.c,v 1.4 1993/04/15 03:35:13 mcooper Exp mcooper $";
#endif

/*
 * Media Access Control (MAC) info routines
 */

#include "defs.h"

#if	GETMAC_TYPE == GETMAC_NIT

#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <net/if.h>
#include <net/nit_if.h>

/*
 * Find and set the MAC info using the Network Interface Tap (NIT)
 */
extern void SetMacInfo(DevName, NetIf)
    char 		       *DevName;
    NETIF 		       *NetIf;
{
    register struct sockaddr   *SockAddr;
    struct ifreq 	        ifreq;
    char 		       *ether_ntoa(), Buf[MAXHOSTNAMLEN+1];
    int 		        Desc;

    if (!NetIf)
	return;

    if ((Desc = open("/dev/nit", O_RDONLY)) == SYSFAIL) {
	if (Debug) Error("open /dev/nit failed");
	return;
    }

    /*
     * Bind to NIT for DevName
     */
    strncpy(ifreq.ifr_name, DevName, sizeof ifreq.ifr_name);
    if (ioctl(Desc, NIOCBIND, (caddr_t) &ifreq) < 0) {
	if (Debug) Error("ioctl:  NIOCBIND");
	return;
    }

    /*
     * Get address
     */
    if (ioctl(Desc, SIOCGIFADDR, (caddr_t)&ifreq) < 0) {
	if (Debug) Error("ioctl (SIOCGIFADDR)");
	return;
    }

    (void) close(Desc);

    SockAddr = (struct sockaddr *)&ifreq.ifr_addr;
    NetIf->ni_macaddr = strdup(ether_ntoa((struct ether_addr *) 
					  SockAddr->sa_data));

    if (ether_ntohost(Buf, (struct ether_addr *) SockAddr->sa_data) == 0)
	NetIf->ni_macname = strdup(Buf);
}
#endif	/* HAVE_NIT */

#if	GETMAC_TYPE == GETMAC_DLPI

#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <net/if.h>
#include <sys/dlpi.h>
#include <netinet/in.h>
#include <netinet/if_ether.h>
#include "dlpi.h"

/*
 * Find and set the MAC info using the Data Link Provider Interface (DLPI)
 */
extern void SetMacInfo(DevName, NetIf, Device)
    char 		       *DevName;
    NETIF 		       *NetIf;
    DEVICE		       *Device;
{
    char 		        buff[MAXHOSTNAMLEN+1];
    int 		        fd;
    long			dlbuf[MAXDLBUF];
    static char			devname[MAXPATHLEN];
    static struct ether_addr	ether_addr;
    register char	       *cp;
    union DL_primitives	       *dlp;
    u_char			addr[MAXDLADDR];

    if (!NetIf)
	return;

    dlp = (union DL_primitives *) dlbuf;
    (void) sprintf(devname, "%s/%s", _PATH_DEV, DevName);
    /*
     * Remove unit part of name from the device name.
     */
    if (cp = rindex(devname, '/')) {
	++cp;
	while (!isdigit(*cp) && ++cp);
	if (isdigit(*cp))
	    *cp = C_NULL;
    }

    if ((fd = open(devname, O_RDWR)) == SYSFAIL) {
	if (Debug) Error("Cannot open %s: %s.", devname, SYSERR);
	return;
    }

    /*
     * Setup
     */
    dlattachreq(fd, Device->dv_unit);
    dlokack(fd, dlbuf);
    dlbindreq(fd, 0, 0, DL_CLDLS, 0, 0);
    dlbindack(fd, dlbuf);

    /*
     * Get current physical address
     */
    dlphysaddrreq(fd, DL_CURR_PHYS_ADDR);
    dlphysaddrack(fd, dlbuf);

    addrtostring(OFFADDR(dlp, dlp->physaddr_ack.dl_addr_offset),
		 dlp->physaddr_ack.dl_addr_length, addr);

    NetIf->ni_macaddr = strdup(addr);

    BCOPY((char *) OFFADDR(dlp, dlp->physaddr_ack.dl_addr_offset), 
	  (char *) ether_addr.ether_addr_octet, 
	  dlp->physaddr_ack.dl_addr_length);
    if (ether_ntohost(buff, &ether_addr) == 0)
	NetIf->ni_macname = strdup(buff);

    /*
     * Get factory physical address
     */
    dlphysaddrreq(fd, DL_FACT_PHYS_ADDR);
    dlphysaddrack(fd, dlbuf);

    addrtostring(OFFADDR(dlp, dlp->physaddr_ack.dl_addr_offset),
		 dlp->physaddr_ack.dl_addr_length, addr);

    NetIf->ni_fmacaddr = strdup(addr);

    BCOPY((char *) OFFADDR(dlp, dlp->physaddr_ack.dl_addr_offset), 
	  (char *) ether_addr.ether_addr_octet, 
	  dlp->physaddr_ack.dl_addr_length);
    if (ether_ntohost(buff, &ether_addr) == 0)
	NetIf->ni_fmacname = strdup(buff);

    /*
     * We're done
     */
    dlunbindreq(fd);
    dldetachreq(fd);
    (void) close(fd);
}
#endif	/* HAVE_DLPI */

#if	defined(HAVE_PACKETFILTER)

#include <sys/time.h>
#include <net/pfilt.h>

#include <fcntl.h>
#include <sys/socket.h>
#include <net/if.h>
#include <netinet/in.h>
#include <netinet/if_ether.h>

#if	defined(NEED_ETHER_ADDR)
/*
 * This didn't appear in <netinet/if_ether.h> until Ultrix 4.2 (4.1?)
 */
struct ether_addr {
	u_char	ether_addr_octet[6];
};
#endif	/* NEED_ETHER_ADDR */

/*
 * Get network type information
 */
static char *GetNetType(type)
    int				type;
{
    extern NAMETAB		NetTypes[];
    register int		i;

    for (i = 0; NetTypes[i].name; i++)
	if (NetTypes[i].value == type)
	    return(NetTypes[i].name);

    return((char *) NULL);
}

/*
 * Find and set the MAC info using the Packet Filter
 */
extern void SetMacInfo(DevName, Netif, Device)
     char 		       *DevName;
     NETIF 		       *Netif;
     DEVICE		       *Device;
{
    struct endevp		endevp;
    struct ether_addr		ether_addr;
    char 		       *ether_ntoa(), HostBuf[MAXHOSTNAMLEN+1];
    char 		       *p;
    int 		        Desc;

    if (!DevName || !Netif)
	return;

    /*
     * Open this device using the packet filter
     */
    if ((Desc = pfopen(DevName, O_RDONLY)) < 0) {
	if (Debug) Error("pfopen %s failed: %s.", DevName, SYSERR);
	return;
    }

    /*
     * Retrieve info
     */
    if (ioctl(Desc, EIOCDEVP, &endevp) < 0) {
	if (Debug) Error("ioctl EIOCDEVP of %s failed: %s.", DevName, SYSERR);
	return;
    }

    close(Desc);

    /*
     * Convert address into ethers(5) format
     */
    BCOPY((char *) endevp.end_addr,
	  (char *) ether_addr.ether_addr_octet,
	  endevp.end_addr_len);

    /*
     * Set what we now know.
     */
    if (p = ether_ntoa(&ether_addr))
	Netif->ni_macaddr = strdup(p);

    if (ether_ntohost(HostBuf, &ether_addr) == 0)
	Netif->ni_macname = strdup(HostBuf);

    if (Device && (p = GetNetType(endevp.end_dev_type)))
	Device->dv_desc = p;
}
#endif	/* HAVE_PACKETFILTER */

#if	!defined(GETMAC_TYPE)
extern void SetMacInfo(DevName, NetIf)
    /*ARGSUSED*/
    char 		       *DevName;
    NETIF 		       *NetIf;
{
    /* Do Nothing */
}
#endif	/* !GETMAC_TYPE */
