/*
 * Copyright (c) 1992-1993 Michael A. Cooper.
 * This software may be freely distributed provided it is not sold for 
 * profit and the author is credited appropriately.
 */

#ifndef lint
static char *RCSid = "$Header: /src/common/usc/bin/sysinfo/RCS/memory.c,v 1.14 1993/06/08 18:15:07 mcooper Exp $";
#endif


/*
 * Memory related functions.
 */

#include "defs.h"

/*
 * Divide and Round Up
 */
extern int DivRndUp(Num, Div)
    unsigned long 		Num;
    unsigned long 		Div;
{
    int 			i;

    i = Num / Div;

    return((Num % Div) ? i+1 : i);
}

#if	PHYSMEM_TYPE == PHYSMEM_PHYSMEM

#include <fcntl.h>
#include <nlist.h>

static char PhysmemSYM[] = 	"_physmem";

/*
 * Common method of determining amount of physical memory in a
 * BSD Unix machine.
 *
 * Get memory by reading the variable "physmem" from the kernel
 * and the system page size.
 */ 
static int GetMemoryAmount()
{
    struct nlist	       *nlptr;
    unsigned long 		Bytes;
    int		 		Amount = -1;
    kvm_t		       *kd;
    int				physmem;

    if (kd = KVMopen()) {
	if ((nlptr = KVMnlist(kd, PhysmemSYM, (struct nlist *)NULL)) == NULL)
	    return(-1);

	if (CheckNlist(nlptr))
	    return(-1);

	if (KVMread(kd, nlptr->n_value, (char *)&physmem, 
		    sizeof(physmem)) >= 0) {
	    /*
	     * Could use ctob() instead of "Page Size * Num Pages",
	     * but this is more portable.
	     */
	    Bytes = (unsigned long) (getpagesize() * physmem);
	    if (Debug)
		printf("Bytes = %d physmem = %d pagesize = %d\n", 
		       Bytes, physmem, getpagesize());
	    Amount = DivRndUp(Bytes, MBYTES);
	}
    }

    if (kd)
	KVMclose(kd);

    return(Amount);
}
#endif	/* PHYSMEM_PHYSMEM */

#if	!defined(PHYSMEM_TYPE)
static int GetMemoryAmount()
{
    return(0);
}
#endif	/* !PHYSMEM_TYPE */

/*
 * Get amount of physical memory
 */
extern char *GetMemory()
{
    static char			Buff[BUFSIZ];
    int				Amount;

    Amount = GetMemoryAmount();
    if (Amount > 0) {
	(void) sprintf(Buff, "%d MB", Amount);
	return(Buff);
    }

    return((char *) NULL);
}
