/*
 * Copyright (c) 1992-1993 Michael A. Cooper.
 * This software may be freely distributed provided it is not sold for 
 * profit and the author is credited appropriately.
 */

#ifndef lint
static char *RCSid = "$Header: /src/common/usc/bin/sysinfo/RCS/os-sunos4.c,v 1.6 1993/04/09 01:17:34 mcooper Exp mcooper $";
#endif

/*
 * SunOS 4.x specific routines
 */

#include "defs.h"
#include <nlist.h>
#include <sys/stat.h>
#include <sys/mtio.h>

/*
 * Characters for disk partitions
 */
char 			PartChars[] = "abcdefgh";

/*
 * CPU (model) symbol
 */
char 			CpuSYM[] = "_cpu";

/*
 * Name of generic magnetic tape device.
 */
#define MTNAME		"mt"

/*
 * Generally used variables
 */
static kvm_t 		       *kd = NULL;
static struct stat 		StatBuf;
static DEVICE 		       *Device;
static char 			Buf[BUFSIZ];
extern char		        RomVecSYM[];
struct dkctlrtab	       *GetDkCtlrTab();
NAMETAB			       *GetFBTab();
static int			OpenPROMTraverse();

#if	defined(HAVE_MAINBUS)
/*
 * Build a device tree by searching the MainBus
 */

#define DV_SIZE	(sizeof(struct mb_device))
#define DR_SIZE (sizeof(struct mb_driver))
extern char		 	MainBusSYM[];

/*
 * Build device tree by looking at mainbus (mb) devices
 */
extern int BuildMainBus(TreePtr)
    DEVICE 		       **TreePtr;
{
    static struct nlist		nlistbuf;
    struct nlist	       *nlptr;
    static struct mb_device 	Device;
    static struct mb_driver 	Driver;
    static char 		CtlrName[BUFSIZ], DevName[BUFSIZ];
    static DEVDATA 		DevData;
    u_long 			Addr, DeviceAddr;
    DEVICE 		       *Dev;

    /*
     * Read table address from kernel
     */
    if (!(kd = KVMopen()))
	return(-1);

    if ((nlptr = KVMnlist(kd, MainBusSYM, &nlistbuf)) == NULL)
	return(-1);

    if (CheckNlist(nlptr))
	return(-1);

    /*
     * Read each device table entry.  A NULL device.mb_driver
     * indicates that we're at the end of the table.
     */
    for (DeviceAddr = nlptr->n_value; DeviceAddr; 
	 DeviceAddr += DV_SIZE) {

	/*
	 * Read this device
	 */
	if (KVMread(kd, DeviceAddr, (char *) &Device, DV_SIZE)) {
	    if (Debug) 
		Error("Cannot read mainbus device from address 0x%x.", 
		      DeviceAddr);
	    KVMclose(kd);
	    return(-1);
	}

	/*
	 * See if we're done.
	 */
	if (!Device.md_driver)
	    break;

	/*
	 * Read the driver structure
	 */
	Addr = (u_long) Device.md_driver;
	if (KVMread(kd, Addr, (char *) &Driver, DR_SIZE)) {
	    if (Debug) 
		Error("Cannot read driver for mainbus address 0x%x.", Addr);
	    continue;
	}

	/*
	 * Get the device name
	 */
	if (Addr = (u_long) Driver.mdr_dname) {
	    if (KVMread(kd, Addr, (char *) DevName, sizeof(DevName))) {
		if (Debug)
		    Error("Cannot read device name from address 0x%x.", Addr);
		continue;
	    }
	} else
	    DevName[0] = C_NULL;

	/*
	 * Get the controller name
	 * XXX - not if "Device.md_ctlr" is -1; work around botch
	 * in current Auspex releases, where some boards (File Processor,
	 * Primary Memory, etc.) have both a device and a controller name,
	 * despite the fact that there's not both a controller and a
	 * set of 1 or more devices.
	 */
	if ((Addr = (u_long) Driver.mdr_cname) && Device.md_ctlr != -1) {
	    if (KVMread(kd, Addr, (char *) CtlrName, sizeof(CtlrName))) {
		if (Debug)
		    Error("Cannot read controller name from address 0x%x.", 
			  Addr);
		continue;
	    }
	} else
	    CtlrName[0] = C_NULL;

	/* Make sure devdata is clean */
	bzero(&DevData, sizeof(DEVDATA));

	/* Set what we know */
	if (DevName[0]) {
	    DevData.dd_devname = strdup(DevName);
	    DevData.dd_devunit = Device.md_unit;
	}
	if (CtlrName[0]) {
	    DevData.dd_ctlrname = strdup(CtlrName);
	    DevData.dd_ctlrunit = Device.md_ctlr;
	}
	/* 
	 * Mainbus devices such, as SCSI targets, may not exist
	 * but the controller reports them as present
	 */
	if (Device.md_alive)
	    DevData.dd_flags |= DD_MAYBE_ALIVE;

	if (Debug)
	    printf("MainBus: Found \"%s\" (Unit %d) on \"%s\" (Unit %d) %s\n", 
		   DevData.dd_devname, DevData.dd_devunit,
		   DevData.dd_ctlrname, DevData.dd_ctlrunit,
		   (DevData.dd_flags & DD_MAYBE_ALIVE) ? "[MAYBE-ALIVE]" : "");

	/* Probe and add device */
	if (Dev = ProbeDevice(&DevData, TreePtr))
	    AddDevice(Dev, TreePtr);
    }

    KVMclose(kd);
    return(0);
}
#endif	/* HAVE_MAINBUS */

/*
 * Get disk info structure.
 */
static DKinfo *GETdk_info(d, file)
    int 			d;
    char 		       *file;
{
    static DKinfo 		dk_info;

    if (ioctl(d, DKIOCINFO, &dk_info) < 0) {
	if (Debug) Error("%s: DKIOCINFO: %s.", file, SYSERR);
	return(NULL);
    }

    return(&dk_info);
}

/*
 * Get disk configuration structure.
 */
static DKconf *GETdk_conf(d, file)
    int 			d;
    char 		       *file;
{
    static DKconf 		dk_conf;

    if (ioctl(d, DKIOCGCONF, &dk_conf) < 0) {
	if (Debug) Error("%s: DKIOCGCONF: %s.", file, SYSERR);
	return(NULL);
    }

    return(&dk_conf);
}

/*
 * Get disk geometry structure.
 */
static DKgeom *GETdk_geom(d, file)
    int 			d;
    char 		       *file;
{
    static DKgeom 		dk_geom;

    if (ioctl(d, DKIOCGGEOM, &dk_geom) < 0) {
	if (Debug) Error("%s: DKIOCGGEOM: %s.", file, SYSERR);
	return(NULL);
    }

    return(&dk_geom);
}

/*
 * Get disk type structure.
 */
static DKtype *GETdk_type(d, file)
    int 			d;
    char 		       *file;
{
    static DKtype 		dk_type;

    if (ioctl(d, DKIOCGTYPE, &dk_type) < 0) {
	if (errno != ENOTTY)
	    if (Debug) Error("%s: DKIOCGTYPE: %s.", file, SYSERR);
	return(NULL);
    }

    return(&dk_type);
}

/*
 * Check the checksum of a disklabel.
 */
static int DkLblCheckSum(DkLabel)
    DKlabel 		       *DkLabel;
{
    register short 	       *Ptr, Sum = 0;
    register short 		Count;

    Count = (sizeof (DKlabel)) / (sizeof (short));
    Ptr = (short *)DkLabel;

    /*
     * Take the xor of all the half-words in the label.
     */
    while (Count--)
	Sum ^= *Ptr++;

    /*
     * The total should be zero for a correct checksum
     */
    return(Sum);
}

/*
 * Get label information from label on disk.
 * The label is stored in the first sector of the disk.
 * We use the driver specific "read" flag with the DKIOCSCMD
 * ioctl to read the first sector.  There should be a special
 * ioctl to just read the label.
 */
static DKlabel *GETdk_label(d, file, dk_info)
    int 			d;
    char 		       *file;
    DKinfo	 	       *dk_info;
{
    static DKlabel	 	dk_label;
    struct dk_cmd 		dk_cmd;
    struct dkctlrtab 	       *pct;

    if (!file || !dk_info)
	return((DKlabel *) NULL);

    if (!(pct = GetDkCtlrTab((int) dk_info->dki_ctype))) {
	Error("Controller type %d is unknown.", 
	      dk_info->dki_ctype);
	return((DKlabel *) NULL);
    }

    if (pct->ct_rdcmd < 0) {
	if (Debug)
	    Error("Read block on controller type \"%s\" is unsupported.",
		  pct->ct_model);
	return((DKlabel *) NULL);
    }

    bzero((char *) &dk_cmd, sizeof(dk_cmd));
    dk_cmd.dkc_cmd = pct->ct_rdcmd;
    dk_cmd.dkc_flags = DK_SILENT | DK_ISOLATE;
    dk_cmd.dkc_blkno = (daddr_t)0;
    dk_cmd.dkc_secnt = 1;
    dk_cmd.dkc_bufaddr = (char *) &dk_label;
    dk_cmd.dkc_buflen = SECSIZE;

    if (ioctl(d, DKIOCSCMD, &dk_cmd) < 0) {
	if (Debug) Error("%s: DKIOCSCMD: %s.", file, SYSERR);
	return((DKlabel *) NULL);
    }

    if (dk_label.dkl_magic != DKL_MAGIC) {
	Error("%s: Disk not labeled.", file);
	return((DKlabel *) NULL);
    }

    if (DkLblCheckSum(&dk_label)) {
	Error("%s: Bad label checksum.", file);
	return((DKlabel *) NULL);
    }

    return(&dk_label);
}

/*
 * Get the name of a disk (i.e. sd0).
 */
static char *GetDiskName(name, dk_conf, dk_info)
    char 		       *name;
    DKconf 		       *dk_conf;
    DKinfo 		       *dk_info;
{
    if (!dk_conf || !dk_info) {
	if (name)
	    return(name);
	return((char *) NULL);
    }

#if	defined(DKI_HEXUNIT)
    if (FLAGS_ON(dk_info->dki_flags, DKI_HEXUNIT))
	(void) sprintf(Buf, "%s%3.3x", dk_conf->dkc_dname, dk_conf->dkc_unit);
    else
#endif 	/* DKI_HEXUNIT */
	(void) sprintf(Buf, "%s%d", dk_conf->dkc_dname, dk_conf->dkc_unit);

    return(strdup(Buf));
}

/*
 * Get the name of the controller for a disk.
 */
static char *GetDkCtlrName(dk_conf)
    DKconf 		       *dk_conf;
{
    if (!dk_conf)
	return((char *) NULL);

    (void) sprintf(Buf, "%s%d", dk_conf->dkc_cname, dk_conf->dkc_cnum);

    return(strdup(Buf));
}

/*
 * Get the disk controller model name from a disk.
 */
static char *GetDkCtlrModel(dk_info)
    DKinfo	 	       *dk_info;
{
    struct dkctlrtab 	       *pct;

    if (!dk_info)
	return((char *) NULL);

    if (!(pct = GetDkCtlrTab(dk_info->dki_ctype)))
	return(NULL);

    return(pct->ct_model);
}

/*
 * Get a disk controller device from disk info.
 */
static DEVICE *GetDkCtlrDevice(DevData, dk_info, dk_conf)
    DEVDATA 		       *DevData;
    DKinfo	 	       *dk_info;
    DKconf 		       *dk_conf;
{
    DEVICE 		       *MkMasterFromDevData();
    DEVICE 		       *dkctlr;

    if ((dkctlr = NewDevice(NULL)) == NULL)
	return((DEVICE *) NULL);

    bzero((char *) dkctlr, sizeof(*dkctlr));

    dkctlr->dv_type = DT_DISKCTLR;

    /*
     * Get name of controller from devdata if available
     */
    if (DevData && DevData->dd_ctlrname)
	dkctlr = MkMasterFromDevData(DevData);

    if (dk_conf) {
	if (!dkctlr->dv_name) {
	    dkctlr->dv_name = GetDkCtlrName(dk_conf);
	    dkctlr->dv_unit = dk_conf->dkc_cnum;
	}
	dkctlr->dv_addr = dk_conf->dkc_addr;
	dkctlr->dv_prio = dk_conf->dkc_prio;
	dkctlr->dv_vec = dk_conf->dkc_vec;
    }

    if (dk_info)
	dkctlr->dv_model = GetDkCtlrModel(dk_info);

    return(dkctlr);
}

/*
 * Get disk label info from the extracted dk_label info.
 */
static char *GetDiskLabel(dk_label)
    DKlabel 		       *dk_label;
{
    register char 	       *cp;
    char		       *label;

    if (!dk_label)
	return((char *) NULL);

    label = strdup(dk_label->dkl_asciilabel);

    /*
     * The label normally has geometry information in it we don't want
     * to see, so we trim out anything starting with " cyl".
     */
    for (cp = label; cp && *cp; ++cp)
	if (*cp == ' ' && strncasecmp(cp, " cyl", 4) == 0)
	    *cp = C_NULL;

    return(label);
}

/*
 * Get filesystem mount info for a partition.
 */
static char *GetMountInfo(name, part)
    char 		       *name;
    char 		       *part;
{
    static FILE 	       *mf = NULL;
    struct mntent 	       *mntent;
    char 		       *file;

    if (!name)
	return((char *) NULL);

    file = GetCharFile(name, part);

    if (!mf) {
	if ((mf = setmntent(MNTTAB, "r")) == NULL) {
	    Error("%s: Cannot open for reading: %s.", MNTTAB, SYSERR);
	    return(NULL);
	}
    } else
	rewind(mf);

    while (mntent = getmntent(mf))
	if (mntent->mnt_fsname && EQ(mntent->mnt_fsname, file))
	    return(mntent->mnt_dir);

    return((char *) NULL);
}

/*
 * Extract the disk partition info from a disk.
 */
static DISKPART *ExtractDiskPart(name, part, dk_conf, dk_geom)
    char 		       *name;
    char 		       *part;
    DKconf	 	       *dk_conf;
    DKgeom	 	       *dk_geom;
{
    static DISKPART 		diskpart;
    struct dk_map 		dk_map;
    char 		       *file;
    char 		       *p;
    int 			d;

    if (!name || !dk_conf || !dk_geom)
	return((DISKPART *) NULL);

    file = GetRawFile(name, part);

    if (stat(file, &StatBuf) != 0) {
	if (Debug) Error("%s: No such partition.", file);
	return((DISKPART *) NULL);
    }

    if ((d = open(file, O_RDONLY)) < 0) {
	if (Debug)
	    Error("%s: Cannot open for read: %s.", file, SYSERR);
	return((DISKPART *) NULL);
    }

    if (ioctl(d, DKIOCGPART, &dk_map) != 0) {
	Error("%s: Cannot extract partition info: %s.", 
		file, SYSERR);
	return((DISKPART *) NULL);
    }
 
    (void) close(d);

    /*
     * Skip empty partitions
     */
    if (!dk_map.dkl_nblk) {
	if (Debug) Error("%s: partition has no size.", file);
	return((DISKPART *) NULL);
    }

    bzero((char *) &diskpart, sizeof(DISKPART));

    diskpart.dp_name = strdup(part);

    if (p = GetMountInfo(name, part))
	diskpart.dp_mnt = strdup(p);
    /* 
     * If this is the "b" partition on the root device, 
     *  then assume it's swap 
     */
    else if (dk_conf->dkc_unit == 0 && strcmp(part, "b") == 0)
	diskpart.dp_mnt = "swap";

    diskpart.dp_stsect = dk_map.dkl_cylno *
	(dk_geom->dkg_nhead * dk_geom->dkg_nsect);
    diskpart.dp_nsect = dk_map.dkl_nblk;

    return(&diskpart);
}

/*
 * Translate disk partition information from basic
 * extracted disk info.
 */
static DISKPART *GetDiskPart(name, dk_conf, dk_geom)
    char 		       *name;
    DKconf 		       *dk_conf;
    DKgeom 		       *dk_geom;
{
    extern char 		PartChars[];
    register DISKPART 	       *pdp, *dp;
    register int 		i;
    static char 		pname[2];
    DISKPART 		       *base = NULL;

    if (!name || !dk_conf || !dk_geom)
	return((DISKPART *) NULL);

    pname[1] = C_NULL;
    for (i = 0; PartChars[i]; ++i) {
	pname[0] = PartChars[i];
	if (dp = ExtractDiskPart(name, pname, dk_conf, dk_geom)) {
	    if (base) {
		for (pdp = base; pdp && pdp->dp_nxt; pdp = pdp->dp_nxt);
		pdp->dp_nxt = NewDiskPart(dp);
	    } else {
		base = NewDiskPart(dp);
	    }
	}
    }

    return(base);
}

/*
 * Convert all we've learned about a disk to a DEVICE.
 */
static DEVICE *DKtoDiskDrive(name, disktype, DevData,
			     dk_info, dk_label, dk_conf, dk_geom, dk_type)
    char 		       *name;
    int				disktype;
    DEVDATA 		       *DevData;
    DKinfo	 	       *dk_info;
    DKlabel	 	       *dk_label;
    DKconf 		       *dk_conf;
    DKgeom 		       *dk_geom;
    DKtype 		       *dk_type;
{
    DEVICE 		       *Device, *dkctlr;
    DISKDRIVE 		       *diskdrive;
    int				listcnt = 0;

    if ((Device = NewDevice(NULL)) == NULL) {
	Error("Cannot create new device entry.");
	return((DEVICE *) NULL);
    }

    if ((dkctlr = NewDevice(NULL)) == NULL) {
	Error("Cannot create new dkctlr device entry.");
	return((DEVICE *) NULL);
    }

    if ((diskdrive = NewDiskDrive(NULL)) == NULL) {
	Error("Cannot create new diskdrive entry.");
	return((DEVICE *) NULL);
    }

    Device->dv_name 		= GetDiskName(name, dk_conf, dk_info);
    Device->dv_type 		= DT_DISKDRIVE;
    /*
     * Only read partition info we we're going to print it later.
     */
    if (VL_ALL)
	diskdrive->dd_part 	= GetDiskPart(name, dk_conf, dk_geom);
    diskdrive->dd_label 	= GetDiskLabel(dk_label);
    Device->dv_model 		= diskdrive->dd_label;

    if (disktype == DKT_CDROM && diskdrive->dd_label == NULL)
	Device->dv_model 	= "CD-ROM";
	
    if (dk_conf) {
	diskdrive->dd_unit 	= dk_conf->dkc_unit;
	diskdrive->dd_slave 	= dk_conf->dkc_slave;;
    }
    if (dk_geom) {
	diskdrive->dd_dcyl 	= dk_geom->dkg_ncyl;
	diskdrive->dd_pcyl 	= dk_geom->dkg_pcyl;
	diskdrive->dd_acyl 	= dk_geom->dkg_acyl;
	diskdrive->dd_heads 	= dk_geom->dkg_nhead;
	diskdrive->dd_sect 	= dk_geom->dkg_nsect;
	diskdrive->dd_apc 	= dk_geom->dkg_apc;
	diskdrive->dd_rpm 	= dk_geom->dkg_rpm;
	diskdrive->dd_intrlv 	= dk_geom->dkg_intrlv;
    }
    if (dk_type) {
	diskdrive->dd_psect 	= dk_type->dkt_hsect;
	diskdrive->dd_promrev 	= dk_type->dkt_promrev;
    }
    if (dk_info) {
#if	defined(DKI_HEXUNIT)
	if (FLAGS_ON(dk_info->dki_flags, DKI_HEXUNIT))
	    diskdrive->dd_flags |= DF_HEXUNIT;
#endif 	/* DKI_HEXUNIT */
    }
    diskdrive->dd_secsize 	= SECSIZE;

    Device->dv_desclist = (char **) xcalloc(MAXDESCLIST, sizeof(char *));
    Device->dv_desclist[listcnt++] 	= GetDiskCapacity(diskdrive);
    Device->dv_desclist[listcnt] 	= NULL;

    dkctlr 			= GetDkCtlrDevice(DevData, dk_info, dk_conf);

    Device->dv_devspec 		= (caddr_t *) diskdrive;
    Device->dv_master 		= dkctlr;

    return(Device);
}

/*
 * Query and learn about a disk.
 */
extern DEVICE *ProbeDiskDriveGeneric(disktype, name, DevData, DevDataTab)
    /*ARGSUSED*/
    int				disktype;
    char 		       *name;
    DEVDATA 		       *DevData;
    DEVDATATAB 	     	       *DevDataTab;
{
    DEVICE 		       *diskdevice;
    DKinfo 		       *dk_info = NULL;
    DKconf 		       *dk_conf = NULL;
    DKtype 		       *dk_type = NULL;
    DKlabel 		       *dk_label = NULL;
    DKgeom 		       *dk_geom = NULL;
    char 		       *rfile;
    int 			d;

    if (!name)
	return((DEVICE *) NULL);

#if	defined(HAVE_IPI)
    /*
     * XXX - Kludge for IPI "id" disks.
     */
    if (EQ(DevData->dd_devname, "id")) {
	static char		Buf[BUFSIZ];

	(void) sprintf(Buf, "%s%3.3x", 
		       DevData->dd_devname, DevData->dd_devunit);
	name = Buf;
    }
#endif	/* HAVE_IPI */

    if (disktype == DKT_CDROM)
	rfile = GetRawFile(name, NULL);
    else {
	if (stat(rfile = GetRawFile(name, NULL), &StatBuf) != 0)
	    /*
	     * Get the name of the whole disk raw device.
	     */
	    rfile = GetRawFile(name, "c");
    }

    if ((d = open(rfile, O_RDONLY)) < 0) {
	if (Debug) Error("%s: Cannot open for reading: %s.", rfile, SYSERR);
	/*
	 * If we know for sure this drive is present and we
	 * know something about it, then create a minimal device.
	 */
	if ((DevDataTab->ddt_model || DevDataTab->ddt_desc) &&
	    FLAGS_ON(DevData->dd_flags, DD_IS_ALIVE)) {
	    Device = NewDevice((DEVICE *) NULL);
	    Device->dv_name = strdup(name);
	    Device->dv_unit = DevData->dd_devunit;
	    Device->dv_master = MkMasterFromDevData(DevData);
	    Device->dv_type = DT_DISKDRIVE;
	    Device->dv_model = DevDataTab->ddt_model;
	    Device->dv_desc = DevDataTab->ddt_desc;
	    return(Device);
	} else
	    return((DEVICE *) NULL);
    }

    if ((dk_conf = GETdk_conf(d, rfile)) == NULL)
	if (Debug) Error("%s: get dk_conf failed.", rfile);

    if ((dk_info = GETdk_info(d, rfile)) == NULL)
	if (Debug) Error("%s: get dk_info failed.", rfile);

    if ((dk_geom = GETdk_geom(d, rfile)) == NULL)
	if (Debug) Error("%s: get dk_geom failed.", rfile);

    if ((dk_label = GETdk_label(d, rfile, dk_info)) == NULL)
	if (Debug) Error("%s: get dk_label failed.", rfile);

    /*
     * Not all controllers support dk_type
     */
    dk_type = GETdk_type(d, rfile);

    close(d);

    if (!(diskdevice = DKtoDiskDrive(name, disktype, DevData,
				     dk_info, dk_label, 
				     dk_conf, dk_geom, dk_type))) {
	Error("%s: Cannot convert diskdrive information.", name);
	return((DEVICE *) NULL);
    }

    return(diskdevice);
}

/*
 * Probe normal disk drive by calling Generic probe routine.
 */
extern DEVICE *ProbeDiskDrive(name, DevData, DevDataTab)
    char 		       *name;
    DEVDATA 		       *DevData;
    DEVDATATAB 	     	       *DevDataTab;
{
    return(ProbeDiskDriveGeneric(DKT_GENERIC, name, DevData, DevDataTab));
}

/*
 * Probe CDROM disk drive by calling Generic probe routine.
 */
extern DEVICE *ProbeCDROMDrive(name, DevData, DevDataTab)
    char 		       *name;
    DEVDATA 		       *DevData;
    DEVDATATAB 	     	       *DevDataTab;
{
    return(ProbeDiskDriveGeneric(DKT_CDROM, name, DevData, DevDataTab));
}

/*
 * Probe a tape device
 * XXX - this loses if somebody's using the tape, as tapes are exclusive-open
 * devices, and our open will therefore fail.
 * This also loses if there's no tape in the drive, as the open will fail.
 * The above probably applies to most other flavors of UNIX.
 */
extern DEVICE *ProbeTapeDrive(name, DevData, DevDataTab)
     /*ARGSUSED*/
    char 		       *name;
    DEVDATA 		       *DevData;
    DEVDATATAB	 	       *DevDataTab;
{
    extern NAMETAB		MtInfo[];
    DEVICE 		       *Device;
    char 		       *file;
    char 		       *model = NULL;
    char			rfile[BUFSIZ];
    static char 		Buf[BUFSIZ];
    struct mtget 		mtget;
    register int 		i;
    int 			d;

    /*
     * Don't use GetRawFile; that'll just stick an "r" in front of the
     * device name, meaning it'll return the rewind-on-close device.
     * Somebody may have left the tape positioned somewhere other than
     * at the BOT to, for example, write a dump there later in the
     * evening; it'd be rather rude to reposition it out from under them.
     *
     * The above probably applies to most other flavors of UNIX.
     */
    if (!name)
	file = NULL;
    else {
	(void) sprintf(rfile, "/dev/nr%s", name);
	file = rfile;
    }

    if ((d = open(file, O_RDONLY)) < 0) {
	if (Debug)
	    Error("%s Cannot open for read: %s.", file, SYSERR);

	/*
	 * --RECURSE--
	 * If we haven't tried the "mt" name yet, try it now
	 */
	if (strncmp(name, MTNAME, strlen(MTNAME)) != 0) {
	    (void) sprintf(Buf, "%s%d", MTNAME, DevData->dd_devunit);
	    Device = ProbeTapeDrive(Buf, DevData, DevDataTab);
	    if (Device)
		return(Device);
	}

	/*
	 * If we know for sure this drive is present and we
	 * know something about it, then create a minimal device.
	 */
	if ((DevDataTab->ddt_model || DevDataTab->ddt_desc) &&
	    FLAGS_ON(DevData->dd_flags, DD_IS_ALIVE)) {
	    Device = NewDevice((DEVICE *) NULL);
	    /* 
	     * Recreate name from devdata since we might have had to
	     * call ourself with name "rmt?"
	     */
	    (void) sprintf(Buf, "%s%d", DevData->dd_devname, 
			   DevData->dd_devunit);
	    Device->dv_name = strdup(Buf);
	    Device->dv_unit = DevData->dd_devunit;
	    Device->dv_master = MkMasterFromDevData(DevData);
	    Device->dv_type = DT_TAPEDRIVE;
	    Device->dv_model = DevDataTab->ddt_model;
	    Device->dv_desc = DevDataTab->ddt_desc;
	    return(Device);
	} else
	    return((DEVICE *) NULL);
    }

    if (ioctl(d, MTIOCGET, &mtget) != 0) {
	Error("%s: Cannot extract tape status: %s.", file, SYSERR);
	return((DEVICE *) NULL);
    }

    (void) close(d);

    model = "unknown";

    for (i = 0; MtInfo[i].name; ++i) {
	if ((MtInfo[i].value == mtget.mt_type)) {
	    model = MtInfo[i].name;
	    break;
	}
    }

    /*
     * Create and set device info
     */
    Device = NewDevice(NULL);
    Device->dv_name = strdup(name);
    Device->dv_type = DT_TAPEDRIVE;
    if (model)
	Device->dv_model = model;
    else
	Device->dv_model = DevDataTab->ddt_model;
    Device->dv_desc = DevDataTab->ddt_desc;
    Device->dv_unit = DevData->dd_devunit;
    Device->dv_master = MkMasterFromDevData(DevData);

    return(Device);
}

/*
 * Query and learn about a device attached to an Auspex Storage Processor.
 * They'll show up as "ad" in the Mainbus info structure, but that
 * merely reflects the way the slots are set up in the config file.
 * We need to find out what type of device it is at this particular
 * instant (it's subject to change - perhaps even while we're running,
 * but there's not a heck of a lot we can do about that).
 *
 * We do that by trying it as a CD-ROM first, then as a disk, then as
 * a tape; that loses if it's a tape and somebody's using it, but
 * tapes on most if not all UNIX systems can't be probed by us if
 * somebody's using it.
 * The reason why we try it first as a CD-ROM is that if the CD has a
 * partition table, the Auspex driver lets you open the partitions as
 * if it were a disk.
 */
extern DEVICE *ProbeSPDrive(name, DevData, DevDataTab)
     /*ARGSUSED*/
    char 		       *name;
    DEVDATA 		       *DevData;
    DEVDATATAB 	     	       *DevDataTab;
{
    DEVICE 		       *thedevice;
    char			devname[BUFSIZ];

    /*
     * Try it first as a CD-ROM.
     */
    (void) sprintf(devname, "acd%d", DevData->dd_devunit);
    DevData->dd_devname = "acd";
    DevDataTab->ddt_model = "CD-ROM";
    if (thedevice = ProbeCDROMDrive(devname, DevData, DevDataTab))
	return(thedevice);

    /*
     * Not a CD-ROM.  Try a disk.
     */
    (void) sprintf(devname, "ad%d", DevData->dd_devunit);
    DevData->dd_devname = "ad";
    DevDataTab->ddt_model = NULL;
    if (thedevice = ProbeDiskDrive(devname, DevData, DevDataTab))
	return(thedevice);

    /*
     * Not a disk.  Try a tape.
     */
    (void) sprintf(devname, "ast%d", DevData->dd_devunit);
    DevData->dd_devname = "ast";
    if (thedevice = ProbeTapeDrive(devname, DevData, DevDataTab))
	return(thedevice);

    /*
     * None of the above.  Who knows?
     */
    return((DEVICE *) NULL);
}

#if	defined(HAVE_SUNROMVEC)
/*
 * Be backwards compatible with pre-4.1.2 code
 */
#include <mon/sunromvec.h>
#if	defined(OPENPROMS) && !(defined(ROMVEC_VERSION) && \
				(ROMVEC_VERSION == 0 || ROMVEC_VERSION == 1))
#define v_mon_id op_mon_id
#endif
#endif	/* HAVE_SUNROMVEC */

/*
 * Get ROM Version number
 *
 * If "romp" is "defined" (in <mon/sunromvec.h>), then take that
 * as the address of the kernel pointer to "rom" (struct sunromvec).
 * Otherwise, nlist "romp" from the kernel.
 */
extern char *GetRomVer()
{
    static char			RomRev[16];
    struct nlist	       *nlptr;
#if	defined(HAVE_SUNROMVEC)
    static struct sunromvec	Rom;
    register char	       *p;
    register char	       *Addr;
    kvm_t		       *kd;
#if	!defined(romp)
    struct sunromvec	       *romp;

    if (!(kd = KVMopen()))
	return((char *) NULL);

    if ((nlptr = KVMnlist(kd, RomVecSYM, (struct nlist *)NULL)) == NULL)
	return((char *) NULL);

    if (CheckNlist(nlptr))
	return((char *) NULL);

    /*
     * Read the kernel pointer to the sunromvec structure.
     */
    if (KVMread(kd, (u_long) nlptr->n_value, (char *) &romp, 
		sizeof(romp))) {
	if (Debug) Error("Cannot read sunromvec pointer from kernel.");
	return((char *) NULL);
    }

#else	/* romp */

    if (!(kd = KVMopen()))
	return((char *) NULL);

#endif	/* romp */

    /*
     * Read the sunromvec structure from the kernel
     */
    /*SUPPRESS 25*/
    if (KVMread(kd, (u_long) romp, (char *) &Rom, sizeof(struct sunromvec))) {
	if (Debug) Error("Cannot read sunromvec from kernel.");
	return((char *) NULL);
    }

#if	!defined(romp)

    /*
     * XXX Hardcoded values
     */
    (void) sprintf(RomRev, "%d.%d", Rom.v_mon_id >> 16, Rom.v_mon_id & 0xFFFF);

#else	/* romp */

    /*
     * Read the version string from the address indicated by Rom.v_mon_id.
     * Read 1 byte at a time until '\0' is encountered.
     */
    p = RomRev;
    Addr = Rom.v_mon_id;
    do {
	if (KVMread(kd, (u_long) Addr++, p, 1))
	    break;
    } while (p < &RomRev[sizeof(RomRev)-1] && *p++);
    *p = C_NULL;

#endif	/* romp */

    KVMclose(kd);

#endif	/* HAVE_SUNROMVEC */

    return((RomRev[0]) ? RomRev : (char *) NULL);
}
